(* ---------------------------------------------------------------
Title         Q&D Sky
Author        PhG
Overview      some use(d|ful) subroutines for Q&D astro tools
Notes
Bugs
Wish List

--------------------------------------------------------------- *)

IMPLEMENTATION MODULE QD_Sky;

IMPORT Str;
IMPORT MATHLIB;
IMPORT FIO;
IMPORT Lib;
FROM IO IMPORT WrStr,WrLn;
FROM SYSTEM IMPORT ADDRESS,BYTE;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, chkJoker,
isOption, GetOptIndex, GetLongCard, GetString, CharCount, same,
aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
str2048, str16, str1024;

FROM QD_Names IMPORT signtype, sPGMAGIC;

FROM QD_Make IMPORT MAJOR, MINOR;

FROM QD_Calc IMPORT RadiansToDegrees, DegreesToRadians, Mod360;


FROM MATHLIB IMPORT Sin, Cos, ACos;

(* ------------------------------------------------------------ *)

CONST
    blank = " ";
    colon = ":";
    dash  = "-";
    slash = "/";
    dot   = ".";
    coma  = ",";
    hhdelim = "H"; (* uppercase *)
    mmdelim = "M"; (* uppercase *)
    FRsection="france"; (* lowercase *)
    codeParis = "P";
    codeLocal = "L";
    codeEast  = "E";
    codeWest  = "W";
    codeOuest = "O";
    codeUnknown="?";
    codeNorth = "N";
    codeSouth = "S";
CONST
    longitudeParis = -2.3375; (* 0h 09mn 21s Est in degrees *)

(* ---------------------------------------------------------------
   input  : value
   output : int(value) ie int(0.95) gives 0 (no rounding)

   TRUNC(n) gives only a CARDINAL without rounding either
--------------------------------------------------------------- *)

PROCEDURE int (v : LONGREAL) : LONGREAL;
VAR
    tmp : LONGINT; (* this should be enough! *)
BEGIN
    tmp := VAL(LONGINT,v); (* as intended, does NOT perform any rounding *)
    RETURN VAL(LONGREAL,tmp);
END int;

(* ---------------------------------------------------------------
   input  : value
   output : frac(value)

   frac keeps sign
--------------------------------------------------------------- *)

PROCEDURE frac (v : LONGREAL) : LONGREAL;
BEGIN
    RETURN (v - int (v));
END frac;

(* ---------------------------------------------------------------
   input  : value1, value2
   output : value1 mod value2

   mod keeps sign
--------------------------------------------------------------- *)

PROCEDURE mod (v1, v2 : LONGREAL) : LONGREAL;
BEGIN
    RETURN ( MATHLIB.Mod(v1, v2) );
END mod;

(* ---------------------------------------------------------------
   assume day with decimal, and whole date is ok
   valid only for JD >= 0 i.e. NOT before 1.5 Jan -4712 !
   input  : day (with decimals), month, year
   output : JD
   do not forget to add deltaTU if necessary!
--------------------------------------------------------------- *)

PROCEDURE DateToJD (Day,Month,Year:LONGREAL) : LONGREAL;
VAR
    YYYYMMDD  : LONGREAL;
    A,gregorian : LONGREAL; (* gregorian correction *)
    JD        : LONGREAL;
BEGIN
    YYYYMMDD := Year * 10000.0 + Month * 100.0 + Day;
    IF Month < 2.5 THEN
        Year  := Year - 1.0;
        Month := Month + 12.0;
    END;
    IF YYYYMMDD < 15821015.0 THEN (* before 15 october 1582, julian calendar *)
        gregorian := 0.0;
    ELSE
        A         := int (Year / 100.0);
        gregorian := 2.0 - A + int (A / 4.0);
    END;
    JD := int (365.25 * (Year + 4716.0)) + int (30.6001 * (Month + 1.0));
    JD := JD + Day + gregorian - 1524.5;
    RETURN JD;
END DateToJD;

(* ---------------------------------------------------------------
   assume JD >= 0
   input  : JD
   output : day (with decimals), month, year
--------------------------------------------------------------- *)

PROCEDURE JDtoDate (JD : LONGREAL; VAR Day,Month,Year : LONGREAL);
VAR
    Z,F,A,alf,B,C,D,E:LONGREAL;
BEGIN
    JD := JD + 0.5;
    Z := int(JD);
    F := frac(JD);
    IF Z < 2299161.0 THEN
        A := Z;
    ELSE
        alf := int( (Z - 1867216.25) / 36524.25);
        A := Z + 1.0 + alf - int( alf / 4.0 );
    END;
    B := A + 1524.0;
    C := int ( (B - 122.1) / 365.25);
    D := int ( 365.25 * C );
    E := int ( (B - D) / 30.6001 );
    Day := B - D - int( 30.6001 * E) + F;
    IF E < 13.5 THEN
        Month := E - 1.0;
    ELSE
        Month := E - 13.0;
    END;
    IF Month > 2.5 THEN
        Year := C - 4716.0;
    ELSE
        Year := C - 4715.0;
    END;
END JDtoDate;

(* ---------------------------------------------------------------
   find JD at 0h
   input  : JD
   output : JD for this day at 0H
--------------------------------------------------------------- *)

PROCEDURE JDtoJD0H (JD : LONGREAL) : LONGREAL;
VAR
    JD0H : LONGREAL;
BEGIN
    IF frac(JD) < 0.5 THEN
       JD0H := int(JD) - 0.5; (* was < 23H59 so take previous 0H *)
    ELSE
       JD0H := int(JD) + 0.5; (* was 0H or later so take current 0H *)
    END;
    RETURN JD0H;
END JDtoJD0H;

(* ---------------------------------------------------------------
   0=sunday, 1=monday... 6=saturday
   input  : JD
   output : day of week
--------------------------------------------------------------- *)

PROCEDURE JDtoWeekday (JD : LONGREAL) : CARDINAL;
VAR
    v : LONGREAL;
BEGIN
    v := mod ( (JDtoJD0H (JD) + 1.5), 7.0 );
    RETURN ( VAL(CARDINAL, v ) );
END JDtoWeekday;

(* ---------------------------------------------------------------
   input  : JD
   output : T in julian centuries from J2000
--------------------------------------------------------------- *)

PROCEDURE JDtoT (JD : LONGREAL) : LONGREAL;
BEGIN
    RETURN ( (JD - 2451545.0) / 36525.0 );
END JDtoT;

(* ------------------------------------------------------------ *)

PROCEDURE isDigit (c:CHAR):BOOLEAN;
CONST
    digits="0123456789";
BEGIN
    RETURN Belongs(digits,c);
END isDigit;

(*
    specialized for date, time, latitude, longitude check
    check if source s$ and pattern fmt$ match together
    #=any number, ?=any char, [..]=one char from this charset, else exact
    if # or ? needed, use [#] or [?]
*)

PROCEDURE MatchFormat (fmt,S:ARRAY OF CHAR  ) : BOOLEAN;
CONST
    setBegin = "[";
    setEnd   = "]";
    Number   = "#";
    anyChar  = "?";
VAR
    wasdigit : BOOLEAN;
    p        : CARDINAL;
    maxp     : CARDINAL;
    currchar : CHAR;
    i        : CARDINAL;
    maxi     : CARDINAL;
    currfmt  : CHAR;
    pclose   : CARDINAL;
    charset  : str128;
BEGIN
    IF same(S,"") THEN RETURN FALSE; END;
    IF same(fmt,"") THEN RETURN FALSE; END;
    IF CharCount(fmt,setBegin) # CharCount(fmt,setEnd) THEN RETURN FALSE; END;
    (* Q&D checks are done now... *)
    p        := 0;                (* index in source S *)
    maxp     := Str.Length(S)-1;
    wasdigit := FALSE;            (* flag for # pattern *)
    i        := 0;                (* index in pattern fmt *)
    maxi     := Str.Length(fmt)-1;
    LOOP
        currchar := S[p];
        currfmt  := fmt[i];
        CASE currfmt OF
        | anyChar :               (* any char so advance both pointers *)
            INC(i);
            INC(p);
        | Number :                (* any number *)
            CASE isDigit(currchar) OF
            | TRUE :
                wasdigit := TRUE;
                INC(p); (* advance only in source S *)
                IF p > maxp THEN (* if a digit was last char of source S... *)
                    INC(i);      (* ... then advance in fmt, just in case *)
                END;
            | FALSE :
                IF wasdigit=FALSE THEN RETURN FALSE; END; (* no digit at all *)
                wasdigit:=FALSE; (* reset flag *)
                INC (i); (* advance only in pattern fmt and keep new nondigit *)
            ELSE ; (* //ISO fixed silly requirement *) END;
        | setBegin : (* any char in charset [..] *)
            pclose:=Str.NextPos(fmt,setEnd,i+1);
            Str.Slice(charset,fmt,i+1,pclose-(i+1) );
            IF Str.CharPos(charset,currchar)=MAX(CARDINAL) THEN RETURN FALSE; END;
            i:=pclose+1;
            INC(p);
        ELSE (* exact *)
            IF currchar # currfmt THEN RETURN FALSE; END;
            INC (i); (* advance both pointers *)
            INC (p);
        END;
        IF (i > maxi) AND (p > maxp) THEN EXIT; END; (* all is ok, S and fmt match *)
        IF (i > maxi) OR (p > maxp) THEN RETURN FALSE; END; (* chars left in either S or fmt *)
    END;
    RETURN TRUE;
END MatchFormat;

(* ------------------------------------------------------------ *)

PROCEDURE getLastCard (S:ARRAY OF CHAR;VAR n : CARDINAL) : BOOLEAN;
VAR
    v:LONGCARD;
    ok:BOOLEAN;
BEGIN
    v:=Str.StrToCard(S,10,ok);
    IF ok = FALSE THEN RETURN FALSE; END;
    IF v > MAX(CARDINAL) THEN RETURN FALSE; END;
    n := CARDINAL(v);
    RETURN TRUE;
END getLastCard;

PROCEDURE stripCard (VAR S:ARRAY OF CHAR;delim:CHAR;VAR n:CARDINAL) : BOOLEAN;
VAR
    p : CARDINAL;
    R : str128;
    v : LONGCARD;
    ok:BOOLEAN;
BEGIN
    p:=Str.CharPos(S,delim);
    Str.Slice(R,S,0,p);
    Str.Delete(S,0,p+1);
    v:=Str.StrToCard(R,10,ok);
    IF ok = FALSE THEN RETURN FALSE; END;
    IF v > MAX(CARDINAL) THEN RETURN FALSE; END;
    n := CARDINAL(v);
    RETURN TRUE;
END stripCard;

PROCEDURE getCard (S:ARRAY OF CHAR;delim:CHAR;item:CARDINAL;
                   VAR n : CARDINAL) : BOOLEAN;
VAR
    R : str128;
    v:LONGCARD;
    ok:BOOLEAN;
BEGIN
    Str.ItemS(R,S,delim,item);
    v:=Str.StrToCard(R,10,ok);
    IF ok = FALSE THEN RETURN FALSE; END;
    IF v > MAX(CARDINAL) THEN RETURN FALSE; END;
    n := CARDINAL(v);
    RETURN TRUE;
END getCard;

(* dd-mm-yyyy, separator is / : -, and if yyyy < 100, add 1900 *)
PROCEDURE parseDate (unsafe:BOOLEAN;S:ARRAY OF CHAR;VAR day,month,year:CARDINAL) : BOOLEAN;
VAR
    legal : CARDINAL;
BEGIN
    legal := 0;
    UpperCase(S);
    IF MatchFormat("#-#-#",S) THEN INC(legal); END; (* yes, we could use #[-/:.]# *)
    IF MatchFormat("#/#/#",S) THEN INC(legal); END;
    IF MatchFormat("#:#:#",S) THEN INC(legal); END;
    IF MatchFormat("#.#.#",S) THEN INC(legal); END;
    IF legal = 0 THEN RETURN FALSE; END;

    ReplaceChar (S,dash,colon);  (* if separator - *)
    ReplaceChar (S,slash,colon); (* if separator / *)
    ReplaceChar (S,dot,colon);   (* if separator . *)

    IF getCard(S,colon,0,day) = FALSE THEN RETURN FALSE; END;
    IF getCard(S,colon,1,month) = FALSE THEN RETURN FALSE; END;
    IF getCard(S,colon,2,year) = FALSE THEN RETURN FALSE; END;
    IF unsafe = FALSE THEN
        IF year < 100 THEN INC(year,1900); END;
    END;
    IF (day < 1) OR (day > 31) THEN RETURN FALSE; END;
    IF (month < 1) OR (month > 12) THEN RETURN FALSE; END;
    (* this is silly but some may want it ! *)
    IF unsafe THEN RETURN TRUE; END;
    IF (year < minYear) OR (year > maxYear) THEN RETURN FALSE; END;
    RETURN TRUE;
END parseDate;

(* hhHmm or hhH, separator is H or : *)
PROCEDURE parseTime (S:ARRAY OF CHAR;VAR hh,mm:CARDINAL) : BOOLEAN;
VAR
    legal : CARDINAL;
BEGIN
    legal := 0;
    UpperCase(S);
    IF MatchFormat("#H#",S) THEN INC(legal); END; (* yes, we could use #[-/:.]# *)
    IF MatchFormat("#:#",S) THEN INC(legal); END;
    IF MatchFormat("#H" ,S) THEN INC(legal); END;
    IF MatchFormat("#:" ,S) THEN INC(legal); END;
    IF legal = 0 THEN RETURN FALSE; END;

    ReplaceChar (S,hhdelim,colon);  (* if separator H *)

    IF MatchFormat("#:",S) THEN
        Str.Append(S,"0"); (* fix hhH for getCard *)
    END;
    IF getCard(S,colon,0,hh) = FALSE THEN RETURN FALSE; END;
    IF getCard(S,colon,1,mm) = FALSE THEN RETURN FALSE; END;
    IF (hh < 0) OR (hh > 23) THEN RETURN FALSE; END;
    IF (mm < 0) OR (mm > 59) THEN RETURN FALSE; END;
    RETURN TRUE;
END parseTime;

(* ------------------------------------------------------------ *)

(* h m s_ DDD,dd_   DDDMM_   DDD_MM  <E|W|O> *)

PROCEDURE parseLongitude (S:ARRAY OF CHAR;
                          VAR lambdaDeg:LONGREAL;
                          VAR lambdaSec:LONGINT) : BOOLEAN;
CONST
    maxAngle        = 180;
    maxLongitude    = 180.0;
    maxLongitudeSec = 180*60*60;
VAR
    east:BOOLEAN;
    chr : CHAR;
    hh,mm,ss,dd:CARDINAL;
    ok:BOOLEAN;
BEGIN
    UpperCase(S);
    IF Str.CharPos(S,codeEast) # MAX(CARDINAL) THEN
        east:=TRUE;
        chr:=codeEast;
    ELSIF Str.CharPos(S,codeWest) # MAX(CARDINAL) THEN
        east:=FALSE;
        chr:=codeWest;
    ELSIF Str.CharPos(S,codeOuest) # MAX(CARDINAL) THEN
        east:=FALSE;
        chr:=codeOuest;
    ELSE
        chr:="?";
    END;
    IF MatchFormat("#H#M#[EWO]",S) THEN
        IF stripCard(S,"H",hh) = FALSE THEN RETURN FALSE; END;
        IF stripCard(S,"M",mm) = FALSE THEN RETURN FALSE; END;
        IF stripCard(S,chr,ss) = FALSE THEN RETURN FALSE; END;
        IF (hh < 0) OR (hh > 12) THEN RETURN FALSE; END;
        IF (mm < 0) OR (mm > 59) THEN RETURN FALSE; END;
        IF (ss < 0) OR (ss > 59) THEN RETURN FALSE; END;
        lambdaSec:=LONGINT(hh*60+mm)*60+LONGINT(ss); (* seconds of time *)
        lambdaSec:=lambdaSec*15; (* now seconds of arc 24h*60*60 -> 360*60*60 *)
        lambdaDeg:=LONGREAL(lambdaSec)/3600.0;
    ELSIF MatchFormat("#H#[EWO]",S) THEN
        IF stripCard(S,"H",hh) = FALSE THEN RETURN FALSE; END;
        IF stripCard(S,chr,mm) = FALSE THEN RETURN FALSE; END;
        ss := 0;
        IF (hh < 0) OR (hh > 12) THEN RETURN FALSE; END;
        IF (mm < 0) OR (mm > 59) THEN RETURN FALSE; END;
        IF (ss < 0) OR (ss > 59) THEN RETURN FALSE; END;
        lambdaSec:=LONGINT(hh*60+mm)*60+LONGINT(ss);
        lambdaSec:=lambdaSec*15;
        lambdaDeg:=LONGREAL(lambdaSec)/3600.0;
    ELSIF MatchFormat("#H[EWO]",S) THEN
        IF stripCard(S,"H",hh) = FALSE THEN RETURN FALSE; END;
        mm := 0;
        ss := 0;
        IF (hh < 0) OR (hh > 12) THEN RETURN FALSE; END;
        IF (mm < 0) OR (mm > 59) THEN RETURN FALSE; END;
        IF (ss < 0) OR (ss > 59) THEN RETURN FALSE; END;
        lambdaSec:=LONGINT(hh*60+mm)*60+LONGINT(ss);
        lambdaSec:=lambdaSec*15;
        lambdaDeg:=LONGREAL(lambdaSec)/3600.0;
    ELSIF MatchFormat("#,#[EWO]"  ,S) THEN
        Str.Subst(S,coma,dot);
        Str.Subst(S,chr,"");
        lambdaDeg:=Str.StrToReal(S,ok);
        IF ok = FALSE THEN RETURN FALSE; END;
        lambdaSec:=LONGINT(lambdaDeg*3600.0+0.5);       (* round it ! *)
    ELSIF MatchFormat("##[EWO]"  ,S) THEN
        IF stripCard(S,"",dd) = FALSE THEN RETURN FALSE; END;
        IF stripCard(S,chr,mm) = FALSE THEN RETURN FALSE; END;
        IF (dd < 0) OR (dd > maxAngle) THEN RETURN FALSE; END;
        IF (mm < 0) OR (mm > 59) THEN RETURN FALSE; END;
        lambdaSec:=LONGINT(dd*60+mm)*60;
        lambdaDeg:=LONGREAL(lambdaSec)/3600.0;
    ELSIF MatchFormat("#[EWO]"   ,S) THEN
        IF stripCard(S,"",dd) = FALSE THEN RETURN FALSE; END;
        mm:=0;
        IF (dd < 0) OR (dd > maxAngle) THEN RETURN FALSE; END;
        IF (mm < 0) OR (mm > 59) THEN RETURN FALSE; END;
        lambdaSec:=LONGINT(dd*60+mm)*60;
        lambdaDeg:=LONGREAL(lambdaSec)/3600.0;
    ELSIF MatchFormat("#[EWO]#"   ,S) THEN
        IF stripCard(S,chr,dd) = FALSE THEN RETURN FALSE; END;
        IF getLastCard(S,mm) = FALSE THEN RETURN FALSE; END;
        IF (dd < 0) OR (dd > maxAngle) THEN RETURN FALSE; END;
        IF (mm < 0) OR (mm > 59) THEN RETURN FALSE; END;
        lambdaSec:=LONGINT(dd*60+mm)*60;
        lambdaDeg:=LONGREAL(lambdaSec)/3600.0;
    ELSE
        RETURN FALSE;
    END;
    IF lambdaDeg > maxLongitude THEN RETURN FALSE; END;
    IF lambdaSec > maxLongitudeSec THEN RETURN FALSE; END;
    IF east=TRUE THEN
        lambdaDeg:=-lambdaDeg;
        lambdaSec:=-lambdaSec;
    END;
    RETURN TRUE;
END parseLongitude;

(* ------------------------------------------------------------ *)

PROCEDURE DateTimeToJD (day,month,year,hh,mm:CARDINAL) : LONGREAL;
CONST
    MinutesPerHour = 60.0;
    HoursPerDay    = 24.0;
VAR
    d : LONGREAL;
BEGIN
    d := LONGREAL(day)+(LONGREAL(hh)+LONGREAL(mm)/MinutesPerHour)/HoursPerDay;
    RETURN DateToJD(d,LONGREAL(month),LONGREAL(year));
END DateTimeToJD;

(* ------------------------------------------------------------ *)

(*
         du (T.U)      au (T.U.)  HL=T.U.plus val, P ou L
   parse "1911:11-03,00/1916:14-06,23=0"
         "aaaa:jj-mm,hh/aaaa:jj-mm,hh=c"
         0....'....0....'....0....'....0

*)

(* correctionTU$ is L, P or correction (with sign NOT YET inverted) *)

PROCEDURE CalcBornes (S:ARRAY OF CHAR;lambdaDeg:LONGREAL;
                      VAR borneinf,bornesup:LONGREAL;
                      VAR correctionTU : ARRAY OF CHAR);
VAR
    yi,di,mi,hi : CARDINAL;
    ys,ds,ms,hs : CARDINAL;
    correction  : LONGREAL;
    hh,mm       : CARDINAL;
BEGIN
    Str.Copy(correctionTU,"?"); (* default is PROBLEM ! *)
    IF MatchFormat("#:#-#,#/#:#-#,#=?",S)=FALSE THEN RETURN; END;
    IF stripCard(S,":",yi)=FALSE THEN RETURN; END;
    IF stripCard(S,"-",di)=FALSE THEN RETURN; END;
    IF stripCard(S,",",mi)=FALSE THEN RETURN; END;
    IF stripCard(S,"/",hi)=FALSE THEN RETURN; END;
    IF stripCard(S,":",ys)=FALSE THEN RETURN; END;
    IF stripCard(S,"-",ds)=FALSE THEN RETURN; END;
    IF stripCard(S,",",ms)=FALSE THEN RETURN; END;
    IF stripCard(S,"=",hs)=FALSE THEN RETURN; END;
    UpperCase(S); (* just in case ! L, P or value in hours *)
    IF same(S,codeParis) THEN
        correction:=-longitudeParis / 15.0; (* change sign here, degrees to hours *)
    ELSIF same(S,codeLocal) THEN
        correction:=-lambdaDeg / 15.0; (* change sign here, degrees to hours *)
    ELSIF same(S,codeUnknown) THEN
        correction:=0.0;
    ELSIF MatchFormat("#",S) THEN (* correction format is # *)
        Str.Append(S,"H"); (* force #H *)
        IF parseTime(S,hh,mm)=FALSE THEN RETURN; END;
        correction:=LONGREAL(hh)+LONGREAL(mm)/60.0; (* hours *)
    ELSIF parseTime (S,hh,mm) THEN (* correction format is #h or #h# *)
        correction:=LONGREAL(hh)+LONGREAL(mm)/60.0; (* hours *)
    ELSE
        RETURN;
    END;
    Str.Copy(correctionTU,S);
    correction:=correction / 24.0; (* hours to days *)
    borneinf:=DateTimeToJD(di,mi,yi,hi,0)+correction;
    bornesup:=DateTimeToJD(ds,ms,ys,hs,0)+correction;
END CalcBornes;

PROCEDURE chkZoneFR (day,month,year,hh,mm:CARDINAL;countrysection:ARRAY OF CHAR) : BOOLEAN;
VAR
    dateIbm:LONGREAL;
BEGIN
    IF same(countrysection,FRsection)=FALSE THEN RETURN FALSE; END;
    dateIbm := LONGREAL(mm)/60.0; (* minutes to fraction of hours *)
    dateIbm := LONGREAL(year * 1000000 + month * 10000 + day * 100 + hh) + dateIbm;
    IF (1940050900.0 < dateIbm) AND (dateIbm < 1942110203.0) THEN RETURN TRUE; END;
    RETURN FALSE;
END chkZoneFR;

(* in fact, always L, P or positive # *)

PROCEDURE ToggleSign (VAR S : ARRAY OF CHAR);
BEGIN
    UpperCase(S); (* unecessary... but just in case ! *)
    IF same(S,codeParis) THEN RETURN; END;
    IF same(S,codeLocal) THEN RETURN; END;
    IF same(S,codeUnknown) THEN RETURN; END;
    IF same(S,"0") THEN RETURN; END;
    IF same(S,"0H") THEN RETURN; END;
    CASE S[0] OF
    | "-" :           (* was negative, so DeltaTU is positive *)
        S[0] := "+";  (* we could just delete - sign *)
    | "+" :           (* was positive, so DeltaTU is negative *)
        S[0] := "-";
    ELSE              (* was positive, so DeltaTU is negative *)
        Str.Prepend(S,"-");
    END;
END ToggleSign;

(* ------------------------------------------------------------ *)


(* DDD,dd_   DDDMM_   DDD_MM  <N|S> *)

PROCEDURE parseLatitude (S:ARRAY OF CHAR;
                         VAR phiDeg:LONGREAL;
                         VAR phiSec:LONGINT) : BOOLEAN;
CONST
    maxAngle       = 90;
    maxLatitude    = 90.0;
    maxLatitudeSec = 90*60*60;
VAR
    north:BOOLEAN;
    chr : CHAR;
    dd,mm:CARDINAL;
    ok:BOOLEAN;
BEGIN
    UpperCase(S);
    IF Str.CharPos(S,codeNorth) # MAX(CARDINAL) THEN
        north:=TRUE;
        chr:=codeNorth;
    ELSIF Str.CharPos(S,codeSouth) # MAX(CARDINAL) THEN
        north:=FALSE;
        chr:=codeSouth;
    ELSE
        chr:="?";
    END;
    IF MatchFormat("#,#[NS]"  ,S) THEN
        Str.Subst(S,coma,dot);
        Str.Subst(S,chr,"");
        phiDeg:=Str.StrToReal(S,ok);
        IF ok = FALSE THEN RETURN FALSE; END;
        phiSec:=LONGINT(phiDeg*3600.0+0.5);       (* round it ! *)
    ELSIF MatchFormat("##[NS]"  ,S) THEN
        IF stripCard(S,"",dd) = FALSE THEN RETURN FALSE; END;
        IF stripCard(S,chr,mm) = FALSE THEN RETURN FALSE; END;
        IF (dd < 0) OR (dd > maxAngle) THEN RETURN FALSE; END;
        IF (mm < 0) OR (mm > 59) THEN RETURN FALSE; END;
        phiSec:=LONGINT(dd*60+mm)*60;
        phiDeg:=LONGREAL(phiSec)/3600.0;
    ELSIF MatchFormat("#[NS]"   ,S) THEN
        IF stripCard(S,"",dd) = FALSE THEN RETURN FALSE; END;
        mm:=0;
        IF (dd < 0) OR (dd > maxAngle) THEN RETURN FALSE; END;
        IF (mm < 0) OR (mm > 59) THEN RETURN FALSE; END;
        phiSec:=LONGINT(dd*60+mm)*60;
        phiDeg:=LONGREAL(phiSec)/3600.0;
    ELSIF MatchFormat("#[NS]#"   ,S) THEN
        IF stripCard(S,chr,dd) = FALSE THEN RETURN FALSE; END;
        IF getLastCard(S,mm) = FALSE THEN RETURN FALSE; END;
        IF (dd < 0) OR (dd > maxAngle) THEN RETURN FALSE; END;
        IF (mm < 0) OR (mm > 59) THEN RETURN FALSE; END;
        phiSec:=LONGINT(dd*60+mm)*60;
        phiDeg:=LONGREAL(phiSec)/3600.0;
    ELSE
        RETURN FALSE;
    END;
    IF phiDeg > maxLatitude THEN RETURN FALSE; END;
    IF phiSec > maxLatitudeSec THEN RETURN FALSE; END;
    IF north=FALSE THEN
        phiDeg:=-phiDeg;
        phiSec:=-phiSec;
    END;
    RETURN TRUE;
END parseLatitude;

(* ------------------------------------------------------------ *)

(* L P #H #H# -#H -#H# +#H +#H# *)

PROCEDURE parseDeltaTU (S:ARRAY OF CHAR;
                        VAR deltaDays:LONGREAL;VAR useLocal:BOOLEAN):BOOLEAN;
VAR
    correction:LONGREAL;
    hh,mm:CARDINAL;
BEGIN
    useLocal:=FALSE;
    UpperCase(S);
    IF same(S,codeParis) THEN
        correction:= longitudeParis / 15.0; (* change sign here, degrees to hours *)
    ELSIF same(S,codeLocal) THEN
        useLocal:=TRUE;
        correction:=0.0; (* just in case *)
    ELSIF MatchFormat("#H",S) THEN
        IF parseTime(S,hh,mm)=FALSE THEN RETURN FALSE; END;
        correction:=LONGREAL(hh)+LONGREAL(mm)/60.0; (* hours *)
    ELSIF MatchFormat("#H#",S) THEN
        IF parseTime(S,hh,mm)=FALSE THEN RETURN FALSE; END;
        correction:=LONGREAL(hh)+LONGREAL(mm)/60.0; (* hours *)
    ELSIF MatchFormat("+#H",S) THEN
        Str.Delete(S,0,1);
        IF parseTime(S,hh,mm)=FALSE THEN RETURN FALSE; END;
        correction:=LONGREAL(hh)+LONGREAL(mm)/60.0; (* hours *)
    ELSIF MatchFormat("+#H#",S) THEN
        Str.Delete(S,0,1);
        IF parseTime(S,hh,mm)=FALSE THEN RETURN FALSE; END;
        correction:=LONGREAL(hh)+LONGREAL(mm)/60.0; (* hours *)
    ELSIF MatchFormat("-#H",S) THEN
        Str.Delete(S,0,1);
        IF parseTime(S,hh,mm)=FALSE THEN RETURN FALSE; END;
        correction:=LONGREAL(hh)+LONGREAL(mm)/60.0; (* hours *)
        correction:=-correction;
    ELSIF MatchFormat("-#H#",S) THEN
        Str.Delete(S,0,1);
        IF parseTime(S,hh,mm)=FALSE THEN RETURN FALSE; END;
        correction:=LONGREAL(hh)+LONGREAL(mm)/60.0; (* hours *)
        correction:=-correction;
    ELSE
        RETURN FALSE;
    END;
    deltaDays:=correction / 24.0; (* hours to days *)
    RETURN TRUE;
END parseDeltaTU;

(* ------------------------------------------------------------ *)

PROCEDURE NumToStr (val:CARDINAL;VAR s:ARRAY OF CHAR);
VAR
    ok:BOOLEAN;
BEGIN
    Str.CardToStr(LONGCARD(val),s,10,ok);
END NumToStr;

(* 1.. *)

PROCEDURE GetText (chaine:ARRAY OF CHAR;element:CARDINAL;delim:ARRAY OF CHAR;
                   VAR text:ARRAY OF CHAR);
BEGIN
    Str.Append(chaine,delim); (* useless safety *)
    Str.ItemS(text,chaine,delim,element-1);
END GetText;

PROCEDURE getMonthStr (m:CARDINAL;VAR R:ARRAY OF CHAR);
CONST
    Smois = "janvier,fvrier,mars,avril,mai,juin,"+
            "juillet,aot,septembre,octobre,novembre,dcembre";
BEGIN
    GetText(Smois,m,coma,R);
END getMonthStr;

PROCEDURE DateToStr (d,m,y:CARDINAL; VAR R:ARRAY OF CHAR);
CONST
    sPremier = "1er";
VAR
    s : str80;
BEGIN
    IF d = 1 THEN
        Str.Copy(R,sPremier);
    ELSE
        NumToStr (d,R);
    END;
    Str.Append(R," ");
    getMonthStr(m,s);
    Str.Append(R,s);
    Str.Append(R," ");
    NumToStr(y,s);
    Str.Append(R,s);
END DateToStr;

PROCEDURE TimeToStr (h,m:CARDINAL;VAR R:ARRAY OF CHAR);
CONST
    sMinuit = "minuit";
    sMidi   = "midi";
    sHeure  = " heure";
    sMinute = " minute";
    sPlural = "s";
VAR
    s : str80;
BEGIN
    CASE h OF
    | 0 : Str.Copy(R,sMinuit);
    | 12: Str.Copy(R,sMidi);
    ELSE
        NumToStr(h,R);
        Str.Append(R,sHeure);
        IF h # 1 THEN Str.Append(R,sPlural);END;
    END;
    IF m > 0 THEN
        Str.Append(R," ");
        NumToStr(m,s);
        Str.Append(R,s);
        Str.Append(R,sMinute);
        IF m # 1 THEN Str.Append(R,sPlural);END;
    END;
END TimeToStr;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    degree    = "";
    zero      = "0";
    minus     = "-";
    arcminute = "'";
    strdegree = ""+blank;
    strHHMM   = "h"+blank;
    strMMSS   = "m"+blank; (* this is just for ALPHA ! *)
    strSS     = "s";
    strMMSStime="mn"+blank; (* for time and longitude display *)

VAR                                     (* globerk ! *)
    padchar : CHAR; (* " " or "0" *)
    padplus : CHAR; (* " " or "+" *)
    padcoma : CHAR; (* "," or "." *)

PROCEDURE PadsSetup (padChar,padPlus,padComa:CHAR);
BEGIN
    padchar:=padChar;
    padplus:=padPlus;
    padcoma:=padComa;
END PadsSetup;

(* ------------------------------------------------------------ *)

PROCEDURE doPad (VAR S : ARRAY OF CHAR;
                 v : LONGCARD; digitswanted:CARDINAL; pad:CHAR);
VAR
    ok     : BOOLEAN;
    digits : CARDINAL;
    i      : CARDINAL;
BEGIN
    Str.CardToStr (v,S,10,ok); (* okay, could use  *)
    digits := Str.Length(S);
    INC(digits);
    FOR i := digits TO digitswanted DO
        Str.Prepend(S,pad);
    END;
END doPad;

PROCEDURE fixComa (VAR S : ARRAY OF CHAR;
                   decimals:CARDINAL;pad:CHAR;comaCh:CHAR);
VAR
    digits : CARDINAL;
    p      : CARDINAL;
    i      : CARDINAL;
BEGIN
    digits := Str.Length(S);
    IF decimals >= digits THEN RETURN; END; (* should never happen *)
    p := digits-decimals; (* [1..] *)
    FOR i := p TO digits DO
        IF S[i-1] = pad THEN (* yes we could use isDigit to avoid replacing 0 with 0 but... *)
            S[i-1] := zero; (* avoid ---,-0 *)
        END;
    END;
    Str.Insert(S,comaCh,p); (* at good position : -1 + 1 ! *)
END fixComa;

PROCEDURE fixSign (VAR S : ARRAY OF CHAR;
                   value:LONGREAL;pad:CHAR);
BEGIN
    IF value < 0.0 THEN
        Str.Prepend(S,minus);
    ELSE
        Str.Prepend(S,pad);
    END;
END fixSign;

(* ------------------------------------------------------------ *)

(* values to format MUST be positive ! *)

PROCEDURE doDDDdd (val : LONGREAL; VAR S : ARRAY OF CHAR); (* 0..360 *)
CONST
    Rmul = 100.0;
    max  = 100 * 360;
VAR
    v : LONGCARD;
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    IF v >= max THEN (* 36000 maximum even if 359.99999999 *)
        v := 0;
    END;
    doPad (S,v,5,padchar);      (* ##### *)
    fixComa(S,2,padchar,padcoma);  (* ###,## *)
    Str.Append (S,degree);
END doDDDdd;

PROCEDURE doDDdd (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 100.0;
VAR
    v : LONGCARD;
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    doPad (S,v,4,padchar);      (* #### *)
    fixComa(S,2,padchar,padcoma);  (* ##,## *)
    Str.Append (S,degree);
END doDDdd;

PROCEDURE doDDd (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 10.0;
VAR
    v : LONGCARD;
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    doPad (S,v,3,padchar);      (* ### *)
    fixComa(S,1,padchar,padcoma);  (* ##,# *)
    Str.Append (S,degree);
END doDDd;

PROCEDURE doDDddddd (val : LONGREAL; VAR S : ARRAY OF CHAR); (* for radius vector *)
CONST
    Rmul = 100000.0;
VAR
    v : LONGCARD;
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    doPad (S,v,7,padchar);      (* ####### *)
    fixComa(S,5,padchar,padcoma);  (* ##,##### *)
END doDDddddd;

PROCEDURE doGrDDd (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 10.0;
VAR
    v : LONGCARD;
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    doPad (S,v,3,padchar);      (* ### *)
    fixComa(S,1,padchar,padcoma);  (* ##,# *)
END doGrDDd;

PROCEDURE doGrDDDd (val : LONGREAL; VAR S : ARRAY OF CHAR); (* 0..360 *)
CONST
    Rmul = 10.0;
    max  = 10 * 360;
VAR
    v : LONGCARD;
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    IF v >= max THEN (* 36000 maximum even if 359.99999999 *)
        v := 0;
    END;
    doPad (S,v,4,padchar);      (* #### *)
    fixComa(S,1,padchar,padcoma);  (* ###,# *)
END doGrDDDd;

(* ------------------------------------------------------------ *)

(* degrees / 15 = hours, then hours * 60 * 60 = seconds *)

PROCEDURE doHHMMSS (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 240.0; (* degrees to seconds of time : 360--24*60*60 1--240 *)
    max  = 24 * 60 * 60;
VAR
    v,hh,mm,ss : LONGCARD;
    R : str128; (* oversized ! *)
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5)); (* in seconds of time now *)
    IF v >= max THEN
        v := 0;
    END;
    hh := v DIV (60 * 60);
    mm := (v - hh * 60 * 60) DIV 60;
    ss := (v - hh * 60 * 60) MOD 60;
    doPad (S,hh,2,padchar); (* ## *)
    Str.Append(S,strHHMM);
    doPad (R,mm,2,padchar); (* ## *)
    Str.Append (R,strMMSS);
    Str.Append(S,R);
    doPad (R,ss,2,padchar); (* ## *)
    Str.Append (R,strSS);
    Str.Append (S,R);
END doHHMMSS;

PROCEDURE doHHMMSStime (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 240.0; (* degrees to seconds of time : 360--24*60*60 1--240 *)
    max  = 24 * 60 * 60;
VAR
    v,hh,mm,ss : LONGCARD;
    R : str128; (* oversized ! *)
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5)); (* in seconds of time now *)
    IF v >= max THEN
        v := 0;
    END;
    hh := v DIV (60 * 60);
    mm := (v - hh * 60 * 60) DIV 60;
    ss := (v - hh * 60 * 60) MOD 60;
    doPad (S,hh,2,padchar); (* ## *)
    Str.Append(S,strHHMM);
    doPad (R,mm,2,padchar); (* ## *)
    Str.Append (R,strMMSStime);
    Str.Append(S,R);
    doPad (R,ss,2,padchar); (* ## *)
    Str.Append (R,strSS);
    Str.Append (S,R);
END doHHMMSStime;

(* ------------------------------------------------------------ *)

(* value can be negative *)

PROCEDURE dolatDDdd (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 100.0;
    max  = 100 * 90;
VAR
    v : LONGCARD;
BEGIN
    v := VAL (LONGCARD,(ABS(val) * Rmul + 0.5));
    IF v >= max THEN
        v := max-1;
    END;
    doPad (S,v,4,padchar);      (* #### *)
    fixComa(S,2,padchar,padcoma);  (* ##,## *)
    Str.Append (S,degree);
    fixSign(S,val,padplus);
END dolatDDdd;

PROCEDURE dolatDDd (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 10.0;
    max  = 10 * 90;
VAR
    v : LONGCARD;
BEGIN
    v := VAL (LONGCARD,(ABS(val) * Rmul + 0.5));
    IF v >= max THEN
        v := max-1;
    END;
    doPad (S,v,3,padchar);      (* ### *)
    fixComa(S,1,padchar,padcoma);  (* ##,# *)
    Str.Append (S,degree);
    fixSign(S,val,padplus);
END dolatDDd;

(* ------------------------------------------------------------ *)

(* value MUST be positive *)

PROCEDURE doDDDMM (val : LONGREAL; VAR S : ARRAY OF CHAR); (* 0..360 *)
CONST
    Rmul = 60.0;  (* degrees to minutes of arc *)
    max  = 60 * 360;
VAR
    v : LONGCARD;
    dd,mm : LONGCARD;
    R : str128; (* oversized ! *)
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    IF v >= max THEN
        v := 0;
    END;
    dd := v DIV 60;
    mm := v MOD 60;
    doPad (S,dd,3,padchar); (* ### *)
    Str.Append (S,strdegree);
    doPad (R,mm,2,padchar); (* ## *)
    Str.Append (R,arcminute);
    Str.Append(S,R);
END doDDDMM;

PROCEDURE doDDMM (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 60.0;  (* degrees to minutes of arc *)
    max  = 60 * 90;
VAR
    v : LONGCARD;
    dd,mm : LONGCARD;
    R : str128; (* oversized ! *)
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    IF v >= max THEN
        v := max-1;
    END;
    dd := v DIV 60;
    mm := v MOD 60;
    doPad (S,dd,2,padchar); (* ## *)
    Str.Append (S,strdegree);
    doPad (R,mm,2,padchar); (* ## *)
    Str.Append (R,arcminute);
    Str.Append(S,R);
END doDDMM;

PROCEDURE doGrDDMM (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 60.0;  (* degrees to minutes of arc *)
    max  = 60 * 90;
VAR
    v : LONGCARD;
    dd,mm : LONGCARD;
    R : str128; (* oversized ! *)
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    IF v >= max THEN
        v := max-1;
    END;
    dd := v DIV 60;
    mm := v MOD 60;
    doPad (S,dd,2,padchar); (* ## *)
    Str.Append (S,degree);          (* ch and NOT str here ! *)
    doPad (R,mm,2,padchar); (* ## *)
    Str.Append (R,arcminute);
    Str.Append(S,R);
END doGrDDMM;

PROCEDURE doGrDDDMM (val : LONGREAL; VAR S : ARRAY OF CHAR); (* 0..360 *)
CONST
    Rmul = 60.0;  (* degrees to minutes of arc *)
    max  = 60 * 360;
VAR
    v : LONGCARD;
    dd,mm : LONGCARD;
    R : str128; (* oversized ! *)
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    IF v >= max THEN
        v := 0;
    END;
    dd := v DIV 60;
    mm := v MOD 60;
    doPad (S,dd,3,padchar); (* ### *)
    Str.Append (S,degree); (* ch and NOT str here ! *)
    doPad (R,mm,2,padchar); (* ## *)
    Str.Append (R,arcminute);
    Str.Append(S,R);
END doGrDDDMM;

(* ------------------------------------------------------------ *)

(* value can be negative *)

PROCEDURE dolatDDMM (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 60.0;  (* degrees to minutes of arc *)
    max  = 60 * 90;
VAR
    v : LONGCARD;
    dd,mm : LONGCARD;
    R : str128; (* oversized ! *)
BEGIN
    v := VAL (LONGCARD,(ABS(val) * Rmul + 0.5));
    IF v >= max THEN
        v := max-1;
    END;
    dd := v DIV 60;
    mm := v MOD 60;
    doPad (S,dd,2,padchar); (* ## *)
    Str.Append (S,strdegree);
    doPad (R,mm,2,padchar); (* ## *)
    Str.Append (R,arcminute);
    Str.Append(S,R);
    fixSign(S,val,padplus);
END dolatDDMM;


PROCEDURE doGrlatDDd (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 10.0;
    max  = 10 * 90;
VAR
    v : LONGCARD;
BEGIN
    v := VAL (LONGCARD,(ABS(val) * Rmul + 0.5));
    IF v >= max THEN
        v := max-1;
    END;
    doPad (S,v,3,padchar);      (* ### *)
    fixComa(S,1,padchar,padcoma);  (* ##,# *)
    fixSign(S,val,padplus);
END doGrlatDDd;

PROCEDURE doGrlatDDMM (val : LONGREAL; VAR S : ARRAY OF CHAR);
CONST
    Rmul = 60.0;  (* degrees to minutes of arc *)
    max  = 60 * 90;
VAR
    v : LONGCARD;
    dd,mm : LONGCARD;
    R : str128; (* oversized ! *)
BEGIN
    v := VAL (LONGCARD,(ABS(val) * Rmul + 0.5));
    IF v >= max THEN
        v := max-1;
    END;
    dd := v DIV 60;
    mm := v MOD 60;
    doPad (S,dd,2,padchar); (* ## *)
    Str.Append (S,degree); (* ch and NOT str here ! *)
    doPad (R,mm,2,padchar); (* ## *)
    Str.Append (R,arcminute);
    Str.Append(S,R);
    fixSign(S,val,padplus);
END doGrlatDDMM;

PROCEDURE doGrVectorDDddd (val : LONGREAL; VAR S : ARRAY OF CHAR); (* for radius vector *)
CONST
    Rmul = 1000.0;
VAR
    v : LONGCARD;
BEGIN
    v := VAL (LONGCARD,(val * Rmul + 0.5));
    doPad (S,v,5,padchar);      (* ##### *)
    fixComa(S,3,padchar,padcoma);  (* ##,### *)
END doGrVectorDDddd;

(* ------------------------------------------------------------ *)

PROCEDURE AngleToStr (valDeg:LONGREAL;fmt:displayfmt;
                      VAR S : ARRAY OF CHAR); (* angle must be normalized and in degrees ! *)
BEGIN
    CASE fmt OF
    | DDDdd      : doDDDdd     (valDeg,S); (* ###,##      *)
    | DDdd       : doDDdd      (valDeg,S); (* ##,##       *)
    | DDd        : doDDd       (valDeg,S); (* ##,#        *)
    | DDddddd    : doDDddddd   (valDeg,S); (* ##,#####     *)
    | HHMMSS     : doHHMMSS    (valDeg,S); (* ##h ##m ##s  *)
    | latDDdd    : dolatDDdd   (valDeg,S); (* ##,##      *)
    | latDDd     : dolatDDd    (valDeg,S); (* ##,#       *)
    | DDDMM      : doDDDMM     (valDeg,S); (* ### ##'     *)
    | DDMM       : doDDMM      (valDeg,S); (* ## ##'      *)
    | latDDMM    : dolatDDMM   (valDeg,S); (* ## ##'     *)
    | HHMMSStime : doHHMMSStime(valDeg,S); (* ##h ##mn ##s *)
    | GrDDd      : doGrDDd     (valDeg,S); (* ##,#         *)
    | GrDDMM     : doGrDDMM    (valDeg,S); (* ####'       without space *)
    | GrDDDd     : doGrDDDd    (valDeg,S); (* ###,#        *)
    | GrDDDMM    : doGrDDDMM   (valDeg,S); (* #####'      without space *)
    | GrlatDDd   : doGrlatDDd  (valDeg,S); (* ##,#        *)
    | GrlatDDMM  : doGrlatDDMM (valDeg,S); (* ####'      without space *)
    | GrVectorDDddd:
               doGrVectorDDddd (valDeg,S); (* ##,###       *)
    ELSE ; (* //ISO fixed silly requirement *) END;
END AngleToStr;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*
' checks motion is direct or retrograde
' take 0 boundary into account
' should work for p1 not too far from p2 ( abs(ecart) < 180)
' p1 and p2 are already in degrees and mod360 !!!
' p1 is position at T, p2 is position at T+1
*)


PROCEDURE isDirect (p1,p2 : LONGREAL ) : BOOLEAN;
VAR
    ecart : LONGREAL;
BEGIN
    ecart:=p2-p1;
    IF ABS(ecart) < 180.0 THEN
	IF ecart < 0.0 THEN
	    RETURN FALSE;   (* --- 0 --- p2 -<- p1 --- *)
	ELSE
	    RETURN TRUE;    (* --- 0 --- p1 ->- p2 --- *)
	END;
    ELSE
 	IF ecart < 0.0 THEN
	    RETURN TRUE;    (* --- p1 --- 0 ->- p2 --- *)
	ELSE
	    RETURN FALSE;   (* --- p2 -<- 0 --- p1 --- *)
	END;
    END;
    RETURN TRUE;
END isDirect;

(* ------------------------------------------------------------ *)

PROCEDURE loadTheme (VAR theme:themeFile;file:ARRAY OF CHAR) : BOOLEAN;
VAR
    hnd : FIO.File;
    len : CARDINAL;
    i   : CARDINAL;
    ok  : BOOLEAN;
    adrTheme : ADDRESS;
    ptrTheme : POINTER TO BYTE;
    c        : BYTE;
BEGIN
    hnd := FIO.OpenRead(file);
    IF FIO.Size(hnd) # SIZE(themeFile) THEN
        FIO.Close(hnd);
        RETURN FALSE;
    END;
    len:=FIO.RdBin(hnd,theme,SIZE(themeFile));
    FIO.Close(hnd);
    IF same(theme.ID,themeID) THEN
        IF same(theme.IDEOF,sPGMAGIC) THEN RETURN TRUE; END; (* this was a plain theme *)
    END;
    RETURN FALSE;
END loadTheme;

PROCEDURE saveTheme(theme:themeFile;datafile:ARRAY OF CHAR);
VAR
    hnd : FIO.File;
    i   : CARDINAL;
    ok  : BOOLEAN;
    adrTheme : ADDRESS;
    ptrTheme : POINTER TO BYTE;
    c        : BYTE;
BEGIN
    hnd := FIO.Create(datafile);
    FIO.WrBin(hnd,theme,SIZE(themeFile));
    FIO.Close(hnd);
END saveTheme;

PROCEDURE chkThemeVersion (theme:themeFile) : BOOLEAN;
VAR
    majminref,majmin:CARDINAL; (* maj * 100 + min *)
BEGIN
    RETURN (theme.major = MAJOR); (* we need the same generation *)
(*
    majminref := 100 * MAJOR       + MINOR;
    majmin    := 100 * theme.major + theme.minor;
    IF majminref >= majmin THEN RETURN TRUE; END;
*)
END chkThemeVersion;

(* ------------------------------------------------------------ *)

(*
    warning : for SAFETY, we NEVER round sign to next sign
    so, 29,99 becomes 30,0 or 300' in the correct sign
*)

PROCEDURE AngleToPos (valDeg:LONGREAL;fmt:displayfmt;
                      VAR S : ARRAY OF CHAR; VAR sign:CARDINAL); (* angle must be normalized and in degrees ! *)
CONST
    degPerSign = LONGREAL(30.0);
VAR
    posDeg : LONGREAL;
    v      : LONGCARD;
BEGIN
    posDeg := MATHLIB.Mod (valDeg,degPerSign);
    sign   := CARDINAL (valDeg / degPerSign); (* 0..11 *)

    CASE fmt OF
    | DDdd    :
        (*
        v := VAL (LONGCARD, posDeg * 100.0 + 0.5);
        IF v >= (30 * 100 ) THEN
            posDeg := 0.0;
            INC(sign);
        END;
        *)
        doDDdd    (posDeg,S); (* ##,##      *)
    | DDd     :
        (*
        v := VAL (LONGCARD, posDeg * 10.0 + 0.5);
        IF v >= (30 * 10) THEN
            posDeg := 0.0;
            INC(sign);
        END;
        *)
        doDDd     (posDeg,S); (* ##,#       *)
    | DDMM    :
        (*
        v := VAL (LONGCARD, posDeg * 60.0 + 0.5); (* in minutes now *)
        IF v >= (30 * 60) THEN (* should never happen *)
            posDeg := 0.0;
            INC(sign);
        END;
        *)
        doDDMM    (posDeg,S); (* ## ##'     *)
    | GrDDd  :
        (*
        v := VAL (LONGCARD, posDeg * 10.0 + 0.5);
        IF v >= (30 * 10) THEN
            posDeg := 0.0;
            INC(sign);
        END;
        *)
        doGrDDd     (posDeg,S); (* ##,#       *)
    | GrDDMM  :
        (*
        v := VAL (LONGCARD, posDeg * 60.0 + 0.5); (* in minutes now *)
        IF v >= (30 * 60) THEN (* should never happen *)
            posDeg := 0.0;
            INC(sign);
        END;
        *)
        doGrDDMM    (posDeg,S); (* ####'   without space ! *)
    ELSE ; (* //ISO fixed silly requirement *) END;
    IF sign > ORD(poissons) THEN sign := ORD(belier); END;
END AngleToPos;

(* ------------------------------------------------------------ *)

PROCEDURE fmtLatitude (phiDeg:LONGREAL;VAR sDDMM,sDDdd:ARRAY OF CHAR);
CONST
    sSud = " Sud";
    sNord= " Nord";
VAR
    S : str128;
BEGIN
    IF phiDeg < 0.0 THEN
        S:=sSud;
    ELSE
        S:=sNord;
    END;
    phiDeg:=ABS(phiDeg);
    AngleToStr (phiDeg,DDMM,sDDMM);
    AngleToStr (phiDeg,DDdd,sDDdd);
    Str.Append(sDDMM,S);
    Str.Append(sDDdd,S);
END fmtLatitude;

PROCEDURE fmtLongitude (lambdaDeg:LONGREAL;
                        VAR sHHMMSS,sDDDMM,sDDDdd:ARRAY OF CHAR);
CONST
    sEst  = " Est";
    sOuest= " Ouest";
VAR
    S : str128;
BEGIN
    IF lambdaDeg < 0.0 THEN
        S:=sEst;
    ELSE
        S:=sOuest;
    END;
    lambdaDeg:=ABS(lambdaDeg);
    AngleToStr(lambdaDeg,HHMMSStime,sHHMMSS);
    AngleToStr(lambdaDeg,DDDMM,sDDDMM);
    AngleToStr(lambdaDeg,DDDdd,sDDDdd);
    Str.Append(sHHMMSS,S);

    Str.Append(sDDDMM,S);
    Str.Append(sDDDdd,S);
END fmtLongitude;

PROCEDURE fmtDeltaTU (S:ARRAY OF CHAR;VAR R:ARRAY OF CHAR);
CONST
    sParis = "Mridien de Paris";
    sLocal = "Mridien local";
    sAucune= "Aucune";
    sHeure = " heure";
    sPlus  = "plus ";
    sMoins = "moins ";
VAR
    ok:BOOLEAN;
    hh,mm:CARDINAL;
    plus:BOOLEAN;
    s:str128;
BEGIN
    plus:=TRUE;
    UpperCase(S);
    IF same(S,codeParis) THEN
        Str.Copy(R,sParis);
        RETURN;
    ELSIF same(S,codeLocal) THEN
        Str.Copy(R,sLocal);
        RETURN;
    ELSIF MatchFormat("#H",S) THEN
        ok:=parseTime(S,hh,mm);
    ELSIF MatchFormat("#H#",S) THEN
        ok:=parseTime(S,hh,mm);
    ELSIF MatchFormat("+#H",S) THEN
        Str.Delete(S,0,1);
        ok:=parseTime(S,hh,mm);
    ELSIF MatchFormat("+#H#",S) THEN
        Str.Delete(S,0,1);
        ok:=parseTime(S,hh,mm);
    ELSIF MatchFormat("-#H",S) THEN
        plus:=FALSE;
        Str.Delete(S,0,1);
        ok:=parseTime(S,hh,mm);
    ELSIF MatchFormat("-#H#",S) THEN
        plus:=FALSE;
        Str.Delete(S,0,1);
        ok:=parseTime(S,hh,mm);
    END;
    IF (hh=0) AND (mm=0) THEN
        Str.Copy(R,sAucune);
        RETURN;
    END;

    NumToStr(hh,R);
    Str.Append(R,sHeure);
    IF hh # 1 THEN Str.Append(R,"s");END;
    IF mm > 0 THEN
        Str.Append(R," ");
        NumToStr(mm,s);
        Str.Append(R,s);
        Str.Append(R," minute");
        IF mm # 1 THEN Str.Append(R,"s");END;
    END;
    CASE plus OF
    | TRUE :   s:= sPlus;
    | FALSE :  s:= sMoins;
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Prepend(R,s);
END fmtDeltaTU;

(* ------------------------------------------------------------ *)

(* see QD_CALC.MOD *)

(* warning ! we MAY get 360 instead of 0 ! *)

PROCEDURE localMod360 (VAR Angle : LONGREAL); (* for angles whose range is 0-360 *)
CONST
    circle = LONGREAL(360.0);
    zero   = LONGREAL(0.0);
BEGIN
    Angle := mod (Angle,circle);
    IF Angle < zero THEN
        Angle := Angle + circle;
        Angle := mod (Angle,circle); (* probably useless... *)
(*
    ELSIF Angle >= circle THEN
        Angle := Angle - circle;
        (* Angle := mod (Angle,circle); *)
*)
    END;
END localMod360;

(*
 no longer assume normalized [0..360[ angles : do it again just in case

 check planet in house cuspN, cuspN+1
 return %True if BI <= L < BS

 si bi < bs alors
                 si l < bi alors niet
                 si l < bs alors ok
                 niet
            sinon
                 si l < bs alors ok
                 si l < bi alors niet
                 ok
*)

PROCEDURE between (BorneInf,pos,BorneSup:LONGREAL) : BOOLEAN;
BEGIN
    (* local not to trouble linker with qd_calc.mod Mod360, just in case... *)
    localMod360(BorneInf);
    localMod360(BorneSup);
    localMod360(pos);

    IF BorneInf < BorneSup THEN
	                            (* BI --- BS *)
        IF pos < BorneSup THEN
    	    IF pos < BorneInf THEN
        	RETURN FALSE;	    (* L < BI -- BS *)
            ELSE
                RETURN TRUE; 	    (* ok *)
            END;
        ELSE
    	    RETURN FALSE;           (* BI -- BS <= L *)
        END;
    ELSE
	                            (* BI --- 0 --- BS *)
	IF pos < BorneInf THEN
	    IF pos < BorneSup THEN
        	RETURN TRUE;        (* ok *)
            ELSE
    		RETURN FALSE;       (* BI --- 0 --- BS <= L *)
            END;
        ELSE
    	    RETURN TRUE;	    (* BI <= L --- 0 --- BS *)
        END;
    END;
END between;

(* ------------------------------------------------------------ *)

(* diff# gives the orbe between aspect and deltaP1-P2 for later use *)

PROCEDURE inAspect (lambda1,lambda2:LONGREAL;ecart:CARDINAL;orbe:LONGREAL;
                    VAR diff:LONGREAL) : BOOLEAN;
VAR
    BorneInf,BorneSup : LONGREAL;
    delta : LONGREAL;
BEGIN
    BorneInf:=LONGREAL(ecart)-orbe;
    BorneSup:=LONGREAL(ecart)+orbe;
    IF ecart = 0 THEN BorneInf:=0.0; END; (* for conjonction *)
    delta:=ABS(lambda1-lambda2);
    IF delta > BorneSup THEN delta:=360.0-delta; END;
    diff:= ABS (LONGREAL(ecart) - delta); (* real orbe here *)
    IF (delta < BorneInf) OR (delta > BorneSup) THEN
	RETURN FALSE;
    ELSE
	RETURN TRUE;
    END;
END inAspect;

(* ------------------------------------------------------------ *)

PROCEDURE isAngulaire (l,p0:LONGREAL;ecartmax:CARDINAL;
                       VAR diff:LONGREAL) : BOOLEAN;
VAR
    bi,bs:LONGREAL;
BEGIN
    bi := p0 - LONGREAL(ecartmax);
    bs := p0 + LONGREAL(ecartmax);

    (* fix possible problem when p0 is about 0 *)
    Mod360(bi);
    Mod360(bs);
    Mod360(p0);
    Mod360(l);

    diff:=ABS(l-p0);
    IF diff > bs THEN diff:=360.0-diff; END;
    IF l=bs THEN RETURN TRUE; END; (* include bs limit *)
    RETURN between(bi,l,bs); (* check bi <= l < bs *)
END isAngulaire;

(* ------------------------------------------------------------ *)

PROCEDURE hav (theta:LONGREAL) : LONGREAL;
BEGIN
    RETURN ( ( 1.0-Cos(theta) ) / 2.0 ); (* see Meeus *)
END hav;

PROCEDURE archav (havd:LONGREAL) : LONGREAL;
BEGIN
    RETURN ACos( -(2.0*havd-1.0) );
END archav;

(* use latitude here -- experimental ! see Meeus 16.1 and 16.5 *)

PROCEDURE inAspectBeta (lambda1,beta1,lambda2,beta2:LONGREAL;
                        ecart:CARDINAL;orbe:LONGREAL;
                        VAR diff:LONGREAL) : BOOLEAN;
VAR
    BorneInf,BorneSup : LONGREAL;
    delta : LONGREAL;
    havD : LONGREAL;
BEGIN
    BorneInf:=LONGREAL(ecart)-orbe;
    BorneSup:=LONGREAL(ecart)+orbe;
    IF ecart = 0 THEN BorneInf:=0.0; END; (* for conjonction *)

    DegreesToRadians(lambda1);
    DegreesToRadians(beta1);
    DegreesToRadians(lambda2);
    DegreesToRadians(beta2);
    (* math formula but less precise for small angles : see Meeus *)
    (* delta:=ACos( Sin(beta1)*Sin(beta2)+Cos(beta1)*Cos(beta2)*Cos(lambda1-lambda2) ); *)
    (* same formula preserving accuracy for small angles : see Meeus *)
    havD := hav(beta1-beta2) + Cos(beta1)*Cos(beta2)*hav(lambda1-lambda2);
    delta:= archav(havD);
    RadiansToDegrees(delta);
    IF delta < 0.0 THEN delta:=delta+180.0; END; (* arc COS ! -4 is really 176 etc. *)

    diff:= ABS (LONGREAL(ecart) - delta); (* real orbe here *)
    IF (delta < BorneInf) OR (delta > BorneSup) THEN
	RETURN FALSE;
    ELSE
	RETURN TRUE;
    END;
END inAspectBeta;

(* ------------------------------------------------------------ *)

(*
 p1...p0...p2
 angles dj MOD360 en thorie
 toujours faire delta=bs-bi
                mod360(delta) (i.e. si delta < 0, +360)
                delta * coeff
                recalculer les bornes et refaire mod360 par scurit
*)

PROCEDURE inZone (pos,pi,pm,ps,ki,ks:LONGREAL) : BOOLEAN;
VAR
    delta1,delta2:LONGREAL;
    bi,bs:LONGREAL;
BEGIN
    delta1:= pm-pi; (* toujours bs-bi *)
    delta2:= ps-pm;
    Mod360(delta1);
    Mod360(delta2);
    delta1:= delta1 * ki;
    delta2:= delta2 * ks;
    bi := pm - delta1;
    bs := pm + delta2;
    Mod360(bi);
    Mod360(bs);
    IF pos=bs THEN RETURN TRUE; END; (* include bs limit *)
    RETURN between(bi,pos,bs); (* check bi <= l < bs *)
END inZone;

(* ------------------------------------------------------------ *)

PROCEDURE getMonthAbrev (month:CARDINAL;VAR R:ARRAY OF CHAR);
CONST
    Smonths="Jan,Fv,Mar,Avr,Mai,Jun,Jul,Ao,Sep,Oct,Nov,Dc"; (* was Jui,Jul *)
BEGIN
    Str.ItemS(R,Smonths,coma,month-1);
END getMonthAbrev;

PROCEDURE NumToXX (val:CARDINAL;VAR s:ARRAY OF CHAR);
VAR
    ok:BOOLEAN;
BEGIN
    Str.CardToStr(LONGCARD(val),s,10,ok);
    IF val < 10 THEN Str.Prepend(s,"0"); END; (* always 0 as pad char here *)
END NumToXX;

PROCEDURE shortDateToStr (d,m,y:CARDINAL; abrev:BOOLEAN; VAR R:ARRAY OF CHAR);
VAR
    s : str80;
BEGIN
    NumToXX (d,R);
    Str.Append(R,"-");
    CASE abrev OF
    | FALSE : NumToXX (m,s);
    | TRUE  : getMonthAbrev(m,s);
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Append(R,s);
    Str.Append(R,"-");
    NumToXX (y,s);
    Str.Append(R,s);
END shortDateToStr;

PROCEDURE shortTimeToStr (h,m:CARDINAL;VAR R:ARRAY OF CHAR);
VAR
    s : str80;
BEGIN
    NumToXX(h,R);
    Str.Append(R,":");
    NumToXX(m,s);
    Str.Append(R,s);
END shortTimeToStr;

PROCEDURE shortfmtLatitude (phiDeg:LONGREAL;VAR sDDMM:ARRAY OF CHAR);
CONST
    sS = " S";
    sN = " N";
VAR
    S : str128;
BEGIN
    IF phiDeg < 0.0 THEN
        S:=sS;
    ELSE
        S:=sN;
    END;
    phiDeg:=ABS(phiDeg);
    AngleToStr (phiDeg,DDMM,sDDMM);
    Str.Append(sDDMM,S);
END shortfmtLatitude;

PROCEDURE shortfmtLongitude (lambdaDeg:LONGREAL;
                             VAR sHHMMSS:ARRAY OF CHAR);
CONST
    sE = " E";
    sO = " O";
VAR
    S : str128;
BEGIN
    IF lambdaDeg < 0.0 THEN
        S:=sE;
    ELSE
        S:=sO;
    END;
    lambdaDeg:=ABS(lambdaDeg);
    AngleToStr(lambdaDeg,HHMMSStime,sHHMMSS);
    Str.Append(sHHMMSS,S);
END shortfmtLongitude;

PROCEDURE shortfmtDeltaTU (S:ARRAY OF CHAR;VAR R:ARRAY OF CHAR);
CONST
    sParis = "Mridien de Paris";
    sLocal = "Mridien local";
VAR
    ok:BOOLEAN;
    hh,mm:CARDINAL;
    plus:BOOLEAN;
    s:str128;
BEGIN
    plus:=TRUE;
    UpperCase(S);
    IF same(S,codeParis) THEN
        Str.Copy(R,sParis);
        RETURN;
    ELSIF same(S,codeLocal) THEN
        Str.Copy(R,sLocal);
        RETURN;
    ELSIF MatchFormat("#H",S) THEN
        ok:=parseTime(S,hh,mm);
    ELSIF MatchFormat("#H#",S) THEN
        ok:=parseTime(S,hh,mm);
    ELSIF MatchFormat("+#H",S) THEN
        Str.Delete(S,0,1);
        ok:=parseTime(S,hh,mm);
    ELSIF MatchFormat("+#H#",S) THEN
        Str.Delete(S,0,1);
        ok:=parseTime(S,hh,mm);
    ELSIF MatchFormat("-#H",S) THEN
        plus:=FALSE;
        Str.Delete(S,0,1);
        ok:=parseTime(S,hh,mm);
    ELSIF MatchFormat("-#H#",S) THEN
        plus:=FALSE;
        Str.Delete(S,0,1);
        ok:=parseTime(S,hh,mm);
    END;
    IF (hh=0) AND (mm=0) THEN
        Str.Copy(R,"0H");
        RETURN;
    END;

    NumToStr(hh,R);
    Str.Append(R,"h");
    IF mm > 0 THEN
        Str.Append(R," ");
        NumToXX(mm,s);
        Str.Append(R,s);
        Str.Append(R,"mn");
    END;
    CASE plus OF
    | TRUE :   s:= "+";
    | FALSE :  s:= "-";
    ELSE ; (* //ISO fixed silly requirement *) END;
    Str.Prepend(R,s);
END shortfmtDeltaTU;

(* ------------------------------------------------------------ *)

(* required because of a stupid mistake in RAPPORT and TRANSITS *)

PROCEDURE swapBool (VAR flag1,flag2:BOOLEAN );
VAR
    tmp:BOOLEAN;
BEGIN
    tmp   := flag1;
    flag1 := flag2;
    flag2 := tmp;
END swapBool;

(* ------------------------------------------------------------ *)

(* process {he|she} or {she} in string, assuming everything is legal *)
(* Str.Match is flawed !!! *)

PROCEDURE processGender(VAR S : ARRAY OF CHAR; isWoman : BOOLEAN);
CONST
    openit   = "{";
    closeit  = "}";
    separit  = "|";
CONST
    (* these are the right patterns to look for ! *)
    pattern1 = "*"+openit+"*"+separit+"*"+closeit+"*"; (* "*{*|*}*" *)
    pattern2 = "*"+openit+"*"+closeit+"*";             (* "*{*}*"   *)

VAR
    p1,p2,p3,n : CARDINAL;
BEGIN
    LOOP
        IF Str.Match(S,pattern1) THEN
            p1 := Str.CharPos(S,openit);
            p2 := Str.CharPos(S,separit);
            p3 := Str.CharPos(S,closeit);
            IF isWoman THEN
                n := p2-p1+1;
                Str.Delete(S,p1,n);
                Str.Delete(S,p3-n,1);
            ELSE
                n := p3-p2+1;
                Str.Delete(S,p2,n);
                Str.Delete(S,p1,1);
            END;
        ELSIF Str.Match(S,pattern2) THEN
            p1 := Str.CharPos(S,openit);
            p3 := Str.CharPos(S,closeit);
            IF isWoman THEN
                Str.Delete(S,p3,1);
                Str.Delete(S,p1,1);
            ELSE
                n := p3-p2+1;
                Str.Delete(S,p1,n);
            END;
        ELSE
            EXIT; (* no pattern found *)
        END;
    END;
END processGender;

PROCEDURE refmtMethod (method,ressource:ARRAY OF CHAR;VAR R1,R2:ARRAY OF CHAR);
VAR
    u,d,n,e:str128;
BEGIN
    Str.Copy(R1,method);
    LowerCase(R1);
    R1[0]:=CAP(R1[0]); (* ugly, ugly, ugly ! *)
    Lib.SplitAllPath(ressource,u,d,n,e);
    Str.Concat(R2,n,e);
    LowerCase(R2);
END refmtMethod;

(* assume we're sure of string sizes ! *)

PROCEDURE clearThemeData (VAR t:themeFile  );
CONST
    c = " ";
VAR
    i:CARDINAL;
BEGIN
    FOR i:= 1 TO SIZE(str80) DO
        t.ID    [i-1] := c;
    END;
    FOR i:= 1 TO SIZE(str2) DO
        t.IDEOF [i-1] := c;
    END;
    FOR i:= 1 TO SIZE(str128) DO
        t.odate      [i-1] := c;
        t.otime      [i-1] := c;
        t.odeltaTU   [i-1] := c;
        t.olatitude  [i-1] := c;
        t.olongitude [i-1] := c;
    END;
    FOR i:= 1 TO SIZE(str1024) DO
        t.orenseignements [i-1] := c;
    END;
END clearThemeData;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE stringToVal (S:ARRAY OF CHAR;VAR v:CARDINAL):BOOLEAN;
VAR
    lc:LONGCARD;
    ok:BOOLEAN;
BEGIN
    LtrimBlanks(S); RtrimBlanks(S);
    lc:=Str.StrToCard(S,10,ok);
    IF ok THEN
        IF lc > MAX(CARDINAL) THEN
            ok:=FALSE;
        ELSE
            v:=CARDINAL(lc);
        END;
    END;
    RETURN ok;
END stringToVal;

PROCEDURE greater (a,b:CARDINAL) : CARDINAL;
BEGIN
    IF a < b THEN
        RETURN b;
    ELSE
        RETURN a;
    END;
END greater;

PROCEDURE smaller (a,b:CARDINAL) : CARDINAL;
BEGIN
    IF a < b THEN
        RETURN a;
    ELSE
        RETURN b;
    END;
END smaller;

PROCEDURE calcCircleCoeff (kmul,orbemax:CARDINAL;orbefound:LONGREAL):LONGREAL;
VAR
    v:LONGREAL;
BEGIN
    v:=LONGREAL(orbemax*orbemax)-(orbefound*orbefound);
    IF v < 0.0 THEN
        v:= 0.0;
    ELSE
        v:=LONGREAL(kmul) * MATHLIB.Sqrt(v) / LONGREAL(orbemax);
    END;
    RETURN v;
END calcCircleCoeff;

PROCEDURE card2str (v,digits:CARDINAL):str16;
VAR
    S:str16;
    i:CARDINAL;
    ok:BOOLEAN;
BEGIN
    Str.CardToStr(LONGCARD(v),S,10,ok);
    FOR i:=1 TO digits DO
        IF Str.Length(S) < i THEN Str.Prepend(S,padchar);END;
    END;
    RETURN S;
END card2str;

(* ####,## ##,## *)

PROCEDURE real2str (v:LONGREAL;digits1,digits2:CARDINAL):str16;
VAR
    S:str16;
    ok:BOOLEAN;
    i,wi:CARDINAL;
BEGIN
    Str.FixRealToStr( ABS(v),digits2,S,ok);
    Str.Subst(S,".",padcoma);
    wi:=digits1+1+digits2;
    IF v < 0.0 THEN DEC(wi);END;
    FOR i:=1 TO wi DO
        IF Str.Length(S) < wi THEN Str.Prepend(S,padchar); END;
    END;
    IF v < 0.0 THEN Str.Prepend(S,"-");END;
    RETURN S;
END real2str;

PROCEDURE rangeWarning (unsafe:BOOLEAN;before,after:CARDINAL);
VAR
    i : CARDINAL;
BEGIN
    IF unsafe THEN
        FOR i:=1 TO before DO
            WrLn;
        END;
        WrStr("*************************************************************************");WrLn;
        WrStr("*                                                                       *");WrLn;
        WrStr("* Vous avez spcifi l'option /UNSAFE, pourtant vivement dconseille : *");WrLn;
        WrStr("* vous assumez donc la responsabilit des rsultats au mieux imprcis,  *");WrLn;
        WrStr("* au pire aberrants, obtenus hors de l'intervalle lgal [1600..2200] !  *");WrLn;
        WrStr("*                                                                       *");WrLn;
        WrStr("*************************************************************************");WrLn;
        FOR i:=1 TO after DO
            WrLn;
        END;
    END;
END rangeWarning;


(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* exceptions is used by the two versions of exceptions handling in Windows *)

VAR
    exceptions : CARDINAL; (* 0=irrelevant, 1=initialized, >1=exceptions+1 *)


(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

BEGIN
    PadsSetup (blank,blank,coma);
    exceptions := 0; (* just in case we would FPUalert without a previous FPUfix *)
    FIO.IOcheck := FALSE; (* avoid nasty problems ! *)

END QD_Sky.
